"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express 
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
import kaskada
import kaskada.api.v1alpha.view_pb2 as view_pb

import grpc

def list_views(**kwargs):
    """
    Lists all the views
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = view_pb.ListViewsRequest(**kwargs)
        return client.viewStub.ListViews(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def get_view(**kwargs):
    """
    Get a view by name
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = view_pb.GetViewRequest(**kwargs)
        return client.viewStub.GetView(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)

def create_view(**kwargs):
    """
    Creates a view
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = view_pb.CreateViewRequest(view = view_pb.View(**kwargs))
        return client.viewStub.CreateView(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)

def delete_view(**kwargs):
    """
    Deletes a view
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = view_pb.DeleteViewRequest(**kwargs)
        return client.viewStub.DeleteView(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express 
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
import kaskada
import kaskada.api.v1alpha.compute_pb2 as compute

import grpc

def query(**kwargs):
    """
    Performs a query
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        response_as = kwargs.pop('response_as', None)
        if response_as == 'json_values':
            kwargs['json_values'] = {}
        elif response_as == 'redis_bulk':
            kwargs['redis_bulk'] = {}
        else:
            kwargs['parquet'] = {}
        query_request = compute.QueryRequest(**kwargs)
        return client.computeStub.Query(query_request, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express 
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
import kaskada
import kaskada.api.v1alpha.table_pb2 as table_pb

import datetime
import grpc
import io
import os
import pandas as pd
import random
import requests
import tempfile


def list_tables(**kwargs):
    """
    Lists all the tables
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = table_pb.ListTablesRequest(**kwargs)
        return client.tableStub.ListTables(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)

def get_table(**kwargs):
    """
    Gets a table by name
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = table_pb.GetTableRequest(**kwargs)
        return client.tableStub.GetTable(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)

def create_table(**kwargs):
    """
    Creates a table
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = table_pb.CreateTableRequest(table = table_pb.Table(**kwargs))
        return client.tableStub.CreateTable(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)

def delete_table(**kwargs):
    """
    Deletes a table
    """
    try:
        client = kwargs.pop('client', kaskada.KASKADA_DEFAULT_CLIENT)
        kaskada.validate_client(client)
        req = table_pb.DeleteTableRequest(**kwargs)
        return client.tableStub.DeleteTable(req, metadata=client.metadata)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)

def upload_data(client: Client, table_name: str, f: io.BufferedReader, file_name: str):
    """
    Generic upload data to a table
    """
    try:
        kaskada.validate_client(client)
        upload_url_req = table_pb.GetUploadUrlRequest(
            table_name = table_name,
            file_name = file_name
        )

        upload_url_res = client.tableStub.GetUploadUrl(upload_url_req, metadata=client.metadata)

        upload_url = upload_url_res.url
        headers = {'X-Amz-Meta-Filename': file_name}
        http_response = requests.put(upload_url, headers=headers, data=f)

        # TODO: Determine the return model of this method
        if http_response.status_code == 200:
            return 'Success'
        return 'Unable to upload file. Resulted in status code: {} and message: {}'.format(http_response.status_code, http_response.content)
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)

def upload_file(table_name: str, file_path: str, client=None):
    """
    Uploads a file to a table
    """
    if not client:
        client = kaskada.KASKADA_DEFAULT_CLIENT
    f = open(file_path, 'rb')
    file_name = os.path.basename(f.name)
    return upload_data(client, table_name, f, file_name,)

def upload_dataframe(table_name: str, df: pd.DataFrame, client=None):
    """
    Uploads a dataframe to a table (converts to parquet then uploads)
    """
    if not client:
        client = kaskada.KASKADA_DEFAULT_CLIENT
    with tempfile.TemporaryDirectory() as tmpdirname:
        current_time = datetime.datetime.now().strftime("%Y-%m-%d")
        base_name = "{}_{}.parquet".format(current_time, random.randint(0, 1000))
        file_name = os.path.join(tmpdirname, base_name)
        # write to tempfile in `tmpdirname`
        df.to_parquet(file_name, version="2.0")
        # upload tempfile from `tmpdirname`
        return upload_data(client, table_name, open(file_name, "rb"), base_name)



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express 
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client

import grpc
import os

from grpc_status import rpc_status
from google.rpc import error_details_pb2

KASKADA_DEFAULT_CLIENT = None

def init(**kwags):
    is_demo = kwags.pop('demo', False)
    if is_demo:
        demo_only(**kwags)
    else:
        client(**kwags)

def demo_only(**kwags):
    kwags['client_id'] = kwags.pop('client_id', 'lWYFx0020u4oulh7Z9UB8C5YXVRHNyk4')
    global KASKADA_DEFAULT_CLIENT
    KASKADA_DEFAULT_CLIENT = Client.demo_only(**kwags)
    return KASKADA_DEFAULT_CLIENT

def client(**kwags):
    client_id = kwags.pop('client_id', os.getenv('KASKADA_CLIENT_ID'))
    if client_id == None:
        raise ValueError("Unable to read Kaskada Client ID. Please provide client_id key word argument or set env var KASKADA_CLIENT_ID.")
    client_secret = kwags.pop('client_secret', os.getenv('KASKADA_CLIENT_SECRET'))
    if client_secret == None:
        raise ValueError("Unable to read Kaskada Client Secret. Please provide client_secret key word argument or set env var KASKADA_CLIENT_SECRET.")

    kwags['client_id'] = client_id
    kwags['client_secret'] = client_secret
    
    global KASKADA_DEFAULT_CLIENT
    KASKADA_DEFAULT_CLIENT = Client.authorized(**kwags)
    return KASKADA_DEFAULT_CLIENT

def handleGrpcError(rpc_error: grpc.Call):
    """
    All methods calling `handleGrpcError` do so after detecting a RpcError.  Currently
    all our grpc calls perform a `UnaryUnaryMultiCallable` method, docs linked here:
    https://grpc.github.io/grpc/python/_modules/grpc.html#UnaryUnaryMultiCallable

    Calling a `UnaryUnaryMultiCallable` method results in one of the following:
        Returns:
            The response value for the RPC.

        Raises:
            RpcError: Indicating that the RPC terminated with non-OK status. The
            raised RpcError will also be a grpc.Call for the RPC affording the RPC's
            metadata, status code, and details.

    Methods on a `grpc.Call` object include the following.  Full docs here:
    https://grpc.github.io/grpc/python/_modules/grpc.html#Call

        code(): The StatusCode value for the RPC.
        details(): The details string of the RPC.

    The `grpc_status.rpc_status.from_call(call)` method helps extract details
    objects from `grpc.Call` objects. Docs here: 
    https://grpc.github.io/grpc/python/_modules/grpc_status/rpc_status.html#from_call

        Args:
            call: A grpc.Call instance.

        Returns:
            A google.rpc.status.Status message representing the status of the RPC.

        Raises:
            ValueError: If the gRPC call's code or details are inconsistent with the
            status code and message inside of the google.rpc.status.Status.
    """
    errorMessage = "An error occurred in your request.\n\tError Code: {}\n".format(rpc_error.code().name)
    try:
        status = rpc_status.from_call(rpc_error)
        if status:
            unpacked = []
            errorMessage += "\tError Message: {}\n".format(status.message)
            for detail in status.details:
                try:
                    val = unpack_details(detail)
                    unpacked.append(val)
                except Exception:
                    # maybe report back to wren in the future
                    pass
            if len(unpacked) > 0:
                errorMessage += "\tDetails: {}\n".format(unpacked)
            raise Exception(errorMessage) from None
    except ValueError:
        # maybe report back to wren in the future
        pass
    errorMessage += "\tError Message: {}\n".format(rpc_error.details())
    raise Exception(errorMessage) from None

def unpack_details(grpc_detail):
    """Unpack a grpc status detail field (which is a 'google.protobuf.Any' type).
    `Unpack()` checks the descriptor of the passed-in message object against the stored one 
    and returns False if they don't match and does not attempt any unpacking; True otherwise.
        Source: 
            https://github.com/protocolbuffers/protobuf/blob/master/python/google/protobuf/internal/well_known_types.py#L81
            https://github.com/protocolbuffers/protobuf/blob/master/python/google/protobuf/internal/python_message.py#L1135

        Raises: 
            google.protobuf.message.DecodeError: If it can't deserialize the message object

    `Is()` checks if the `Any` message represents the given protocol buffer type.
    """
    if grpc_detail.Is(error_details_pb2.BadRequest.DESCRIPTOR):
        val = error_details_pb2.BadRequest()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.PreconditionFailure.DESCRIPTOR):
        val = error_details_pb2.PreconditionFailure()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.RetryInfo.DESCRIPTOR):
        val = error_details_pb2.RetryInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.DebugInfo.DESCRIPTOR):
        val = error_details_pb2.DebugInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.QuotaFailure.DESCRIPTOR):
        val = error_details_pb2.QuotaFailure()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.RequestInfo.DESCRIPTOR):
        val = error_details_pb2.RequestInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.ResourceInfo.DESCRIPTOR):
        val = error_details_pb2.ResourceInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.Help.DESCRIPTOR):
        val = error_details_pb2.Help()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.LocalizedMessage.DESCRIPTOR):
        val = error_details_pb2.LocalizedMessage()
        grpc_detail.Unpack(val)
        return val
    else:
        raise ValueError(grpc_detail.type_url)

def handleException(e: Exception):
    raise Exception("An exception occurred: {}".format(e)) from None

def validate_client(client: Client):
    if client == None:
        raise ValueError('Client must be provided')
    if client.computeStub == None:
        raise ValueError('Invalid client provided. Compute service stub was not initialized properly.')
    if client.viewStub == None:
        raise ValueError('Invalid client provided. View service stub was not initialized properly.')
    if client.tableStub == None:
        raise ValueError('Invalid client provided. Table service stubs was not initialized properly.')


"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express 
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from __future__ import print_function
import logging

import grpc
import http.client
import json

import kaskada.api.v1alpha.compute_pb2_grpc as compute_grpc
import kaskada.api.v1alpha.table_pb2_grpc as table_grpc
import kaskada.api.v1alpha.view_pb2_grpc as view_grpc

class Client(object):
    LEGACY_ENGINE = 'LEGACY'

    class AccessToken(object):
        def __init__(self, json_data: str):
            data = json.loads(json_data)
            if 'access_token' not in data and 'expires_in' not in data and 'token_type' not in data:
                raise PermissionError('Unable to validate access token. Token details: {}'.format(json_data))
            self.access_token = data['access_token']
            self.expires_in = data['expires_in']
            self.token_type = data['token_type']

    def getBearerToken(endpoint: str, audience: str, client_id: str, client_secret: str):
        conn = http.client.HTTPSConnection(endpoint)
        payload = {
            'client_id': client_id,
            'client_secret': client_secret,
            'audience': audience, 
            'grant_type': 'client_credentials',
        }
        headers = {
            'content-type': 'application/json',
        }
        conn.request('POST', '/oauth/token', json.dumps(payload), headers)
        res = conn.getresponse()
        data = res.read().decode('utf-8')
        return Client.AccessToken(data)

    def __init__(self, metadata: [(str, str)], **kwags):
        engine = kwags.pop('engine', None)
        endpoint = kwags.pop('endpoint', 'api.kaskada.com:50051')
        is_secure = kwags.pop('is_secure', True)
        if is_secure:
            channel = grpc.secure_channel(endpoint, grpc.ssl_channel_credentials())
        else:
            channel = grpc.insecure_channel(endpoint)
        self.computeStub = compute_grpc.ComputeServiceStub(channel)
        self.tableStub = table_grpc.TableServiceStub(channel)
        self.viewStub = view_grpc.ViewServiceStub(channel)
        self.metadata = metadata
        if engine is not None:
            self.metadata.append(('x-engine', engine))

    def authorized(**kwags):
        credentials_exchange_endpoint = kwags.pop('exchange_endpoint', 'prod-kaskada.us.auth0.com')
        credentials_audience = kwags.pop('audience', 'https://api.prod.kaskada.com')
        client_id = kwags.pop('client_id')
        client_secret = kwags.pop('client_secret')
        token = Client.getBearerToken(credentials_exchange_endpoint, credentials_audience, client_id, client_secret)
        metadata = [('authorization', token.access_token)]
        return Client(metadata, **kwags)

    def demo_only(**kwags):
        client_id = kwags.pop('client_id', 'lWYFx0020u4oulh7Z9UB8C5YXVRHNyk4')
        metadata = [('client-id', client_id)]
        return Client(metadata, **kwags)
